unit varFoo;

interface
uses
  Variants;


// 'Class1'  'Class2'
function CreateELOObject(const ClassName: string): variant;

function varClass1: TVarType;
function varClass2: TVarType;

procedure SetDestFile(const AFileName: string);

implementation

uses
  SysUtils, Classes;

type
  TClass1VariantType = class(TInvokeableVariantType)
  public
    function DoProcedure(const V: TVarData; const Name: string;
      const Arguments: TVarDataArray): Boolean; override;
    procedure Clear(var V: TVarData); override;
    procedure Copy(var Dest: TVarData; const Source: TVarData;
      const Indirect: Boolean); override;
    procedure CastTo(var Dest: TVarData; const Source: TVarData;
      const AVarType: TVarType); override;
  end;

  TClass2VariantType = class(TInvokeableVariantType)
    function DoProcedure(const V: TVarData; const Name: string;
      const Arguments: TVarDataArray): Boolean; override;
    procedure Clear(var V: TVarData); override;
    procedure Copy(var Dest: TVarData; const Source: TVarData;
      const Indirect: Boolean); override;
    procedure CastTo(var Dest: TVarData; const Source: TVarData;
      const AVarType: TVarType); override;
  end;

  TClassData = class(TPersistent)
  private
    FName: string;
  public
    constructor Create(const ClassName: string);
    property AName: string read FName write FName;
    procedure Assign(Source: TPersistent); override;
  end;

  TClassVarData = packed record
    VType: TVarType;
    Reserved1, Reserved2, Reserved3: Word;
    VClassData: TClassData;
    Reserved4: LongInt;
  end;

function CreateELOObject(const ClassName: string): variant;
begin
  varClear(Result);
  if CompareText(ClassName, 'Class1') = 0 then
  begin
    TClassVarData(Result).VType := varClass1;
    TClassVarData(Result).VClassData := TClassData.Create('ClassName1');
  end
  else
    if CompareText(ClassName, 'Class2') = 0 then
    begin
      TClassVarData(Result).VType := varClass2;
      TClassVarData(Result).VClassData := TClassData.Create('ClassName2');
    end
    else
      raise Exception.Create('  !!');
end;

var
  FileName: string;

procedure SetDestFile(const AFileName: string);
begin
  FileName := AFileName;
end;

function SaveToFile(const V: TVarData;
  const Name: string; const Arguments: TVarDataArray): Boolean;
var
  F: TextFile;
  i: integer;
  ObjName, ClassName: string;
  ArgStr: string;
begin
  ClassName := 'Class';
  if TClassVarData(V).VType = varClass1 then
    ClassName := 'Class1';
  if TClassVarData(V).VType = varClass2 then
    ClassName := 'Class2';
  Result := True;
  AssignFile(F, FileName);
  if FileExists(FileName) then
    Append(F)
  else
    Rewrite(F);
  try
    ObjName := TClassVarData(V).VClassData.AName;
    ArgStr := '';
    for i := 0 to High(Arguments) do
      ArgStr := ArgStr + ' ' + VarToStr(Variant(Arguments[i]));
    writeln(F, 'Class:', ClassName, ' Obj:', ObjName, ' Metod:', Name,
      ' Args:', ArgStr);
  finally
    CloseFile(F);
  end;
end;

{ TClass2VariantType }

procedure TClass2VariantType.CastTo(var Dest: TVarData;
  const Source: TVarData; const AVarType: TVarType);
var
  LTemp: TVarData;
begin
  if Source.VType = VarType then
    case AVarType of
      varOleStr:
        VarDataFromOleStr(Dest, TClassVarData(Source).VClassData.AName);
      varString:
        VarDataFromStr(Dest, TClassVarData(Source).VClassData.AName);
    else
      inherited;
    end;
end;

procedure TClass2VariantType.Clear(var V: TVarData);
begin
  V.VType := varEmpty;
  FreeAndNil(TClassVarData(V).VClassData);
end;

procedure TClass2VariantType.Copy(var Dest: TVarData;
  const Source: TVarData; const Indirect: Boolean);
begin
  if Indirect and VarDataIsByRef(Source) then
    VarDataCopyNoInd(Dest, Source)
  else
    with TClassVarData(Dest) do
    begin
      VType := varType;
      VClassData := TClassData.Create('ClassName1');
    end;
end;

function TClass2VariantType.DoProcedure(const V: TVarData;
  const Name: string; const Arguments: TVarDataArray): Boolean;
begin
  Result := False;
  if CompareText(Name, 'FooBar') = 0 then
    Result := SaveToFile(V, Name, Arguments);
end;

{ TClass1VariantType }

procedure TClass1VariantType.CastTo(var Dest: TVarData;
  const Source: TVarData; const AVarType: TVarType);
var
  LTemp: TVarData;
begin
  if Source.VType = VarType then
    case AVarType of
      varOleStr:
        VarDataFromOleStr(Dest, TClassVarData(Source).VClassData.AName);
      varString:
        VarDataFromStr(Dest, TClassVarData(Source).VClassData.AName);
    else
      inherited;
    end;
end;

procedure TClass1VariantType.Clear(var V: TVarData);
begin
  V.VType := varEmpty;
  FreeAndNil(TClassVarData(V).VClassData);
end;

procedure TClass1VariantType.Copy(var Dest: TVarData;
  const Source: TVarData; const Indirect: Boolean);
begin
  if Indirect and VarDataIsByRef(Source) then
    VarDataCopyNoInd(Dest, Source)
  else
    with TClassVarData(Dest) do
    begin
      VType := varType;
      VClassData := TClassData.Create('ClassName2');
    end;
end;

function TClass1VariantType.DoProcedure(const V: TVarData;
  const Name: string; const Arguments: TVarDataArray): Boolean;
begin
  Result := False;
  if (CompareText(Name, 'Foo') = 0) or (CompareText(Name, 'Bar') = 0) then
    Result := SaveToFile(V, Name, Arguments);
end;

var
  Class1VariantType: TClass1VariantType;
  Class2VariantType: TClass2VariantType;

function varClass1: TVarType;
begin
  Result := Class1VariantType.VarType;
end;

function varClass2: TVarType;
begin
  Result := Class2VariantType.VarType;
end;

{ TClassData }

procedure TClassData.Assign(Source: TPersistent);
begin
  if Source is TClassData then
  begin
    FName := (Source as TClassData).AName;
    exit;
  end;
  inherited Assign(Source);
end;

var
  Class1Cnt: integer = 0;
  Class2Cnt: integer = 0;

constructor TClassData.Create(const ClassName: string);
begin
  FName := ClassName;
  if CompareText(ClassName, 'ClassName1') = 0 then
  begin
    inc(Class1Cnt);
    FName := 'First'  + IntToStr(Class1Cnt);
  end;
  if CompareText(ClassName, 'ClassName2') = 0 then
  begin
    inc(Class2Cnt);
    FName := 'Second' + IntToStr(Class2Cnt);
  end;
end;

initialization
  Class1VariantType := TClass1VariantType.Create;
  Class2VariantType := TClass2VariantType.Create;
finalization
  FreeAndNil(Class1VariantType);
  FreeAndNil(Class2VariantType);
end.
